# -*- coding: utf-8 -*-
# Copyright (C) 2018-2023 by the Free Software Foundation, Inc.
#
# This file is part of Postorius.
#
# Postorius is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Postorius is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Postorius.  If not, see <http://www.gnu.org/licenses/>.
#

from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned
from django.db import IntegrityError
from django.http import Http404, HttpResponse, HttpResponseBadRequest
from django.shortcuts import redirect
from django.urls import reverse
from django.views.decorators.http import require_safe
from django.views.generic import CreateView, DeleteView, ListView, UpdateView

from postorius.auth.mixins import DomainOwnerMixin, ListOwnerMixin
from postorius.forms.list_forms import TemplateUpdateForm
from postorius.models import Domain, EmailTemplate, List


def _clean_with_no_strip(field, data):
    """Clean Django's CharField field without strip=True.

    :param field: Instance of the Charfield.
    :param data: The un-cleaned data from the form.
    :returns: Cleaned data.
    """
    field.strip = False
    return field.clean(data)


class ListContextMixin:
    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['list'] = List.objects.get_or_404(
            fqdn_listname=self.kwargs['list_id']
        )
        return context


class ListTemplateIndexView(ListOwnerMixin, ListContextMixin, ListView):

    model = EmailTemplate
    template_name = 'postorius/lists/template_list.html'

    def get_queryset(self):
        return EmailTemplate.objects.filter(identifier=self.kwargs['list_id'])


class ListTemplateCreateView(ListOwnerMixin, ListContextMixin, CreateView):

    template_name = 'postorius/lists/template_add.html'
    model = EmailTemplate
    fields = ['name', 'data']

    def get_success_url(self):
        return reverse('list_template_list', args=(self.kwargs['list_id'],))

    def form_valid(self, form):
        formdata = form.cleaned_data
        formdata['data'] = _clean_with_no_strip(
            form.fields['data'], form.data['data']
        )
        formdata['identifier'] = self.kwargs['list_id']
        formdata['context'] = 'list'
        email_template = EmailTemplate(**formdata)
        # Try to save the model. Some of the unique constraints that we have
        # depend on the mailing_list attribute added above. So, even though
        # we check for form validity, the save can fail.
        try:
            email_template.save()
        except IntegrityError as e:
            form.add_error('name', str(e))
            form.add_error(
                'name',
                'You already have this template set. '
                'Use edit instead of creating a new one.',
            )
            return self.form_invalid(form)
        return redirect(self.get_success_url())


class ListTemplateUpdateView(ListOwnerMixin, ListContextMixin, UpdateView):

    template_name = 'postorius/lists/template_update.html'
    model = EmailTemplate
    form_class = TemplateUpdateForm

    def get_success_url(self):
        return reverse('list_template_list', args=(self.kwargs['list_id'],))


class ListTemplateDeleteView(ListOwnerMixin, ListContextMixin, DeleteView):

    template_name = 'postorius/lists/template_delete.html'
    model = EmailTemplate

    def get_success_url(self):
        return reverse('list_template_list', args=(self.kwargs['list_id'],))


class DomainContextMixin:
    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['domain'] = Domain.objects.get_or_404(
            mail_host=self.kwargs['domain']
        )
        return context


class DomainTemplateIndexView(DomainOwnerMixin, DomainContextMixin, ListView):

    model = EmailTemplate
    template_name = 'postorius/domain/template_index.html'
    header = 'Index'

    def get_queryset(self):
        return EmailTemplate.objects.filter(identifier=self.kwargs['domain'])


class DomainTemplateCreateView(
    DomainOwnerMixin, DomainContextMixin, CreateView
):

    model = EmailTemplate
    template_name = 'postorius/domain/template_add.html'
    fields = ['name', 'data']
    header = 'New Template'

    def get_success_url(self):
        return reverse('domain_template_list', args=(self.kwargs['domain'],))

    def form_valid(self, form):
        formdata = form.cleaned_data
        formdata['data'] = _clean_with_no_strip(
            form.fields['data'], form.data['data']
        )
        formdata['identifier'] = self.kwargs['domain']
        formdata['context'] = 'domain'
        template = EmailTemplate(**formdata)
        # Try to save the model. Some of the unique constraints that we have
        # depend on the mailing_list attribute added above. So, even though we
        # check for form validity, the save can fail.
        try:
            template.save()
        except IntegrityError as e:
            form.add_error('name', str(e))
            form.add_error(
                'name',
                'You already have this template set.'
                'Use edit instead of creating a new one.',
            )
            return self.form_invalid(form)
        return redirect(self.get_success_url())


class DomainTemplateUpdateView(
    DomainOwnerMixin, DomainContextMixin, UpdateView
):

    model = EmailTemplate
    template_name = 'postorius/domain/template_add.html'
    form_class = TemplateUpdateForm
    header = 'Edit Template'

    def get_success_url(self):
        return reverse('domain_template_list', args=(self.kwargs['domain'],))


class DomainTemplateDeleteView(
    DomainOwnerMixin, DomainContextMixin, DeleteView
):

    template_name = 'postorius/domain/template_delete.html'
    model = EmailTemplate

    def get_success_url(self):
        return reverse('domain_template_list', args=(self.kwargs['domain'],))


@require_safe
def get_template_data(request, context, identifier, name):
    # At this point, the request should be authenticated and it's method should
    # be GET. We just need to find the correct template and return it's
    # content, if it exists, return a 404 otherwise.
    if context not in ('list', 'domain'):
        return HttpResponseBadRequest(
            'context should be either "list" or "domain"'
        )
    data = dict(name=name, identifier=identifier, context=context)
    # Depending on the context, populate the right identifier.
    try:
        template = EmailTemplate.objects.get(**data)
    except EmailTemplate.DoesNotExist:
        raise Http404('Template is not defined.')
    except MultipleObjectsReturned:
        raise HttpResponseBadRequest('Multiple Templates exist')

    content_type = 'text/plain'
    if settings.DEFAULT_CHARSET:
        content_type += '; charset=' + settings.DEFAULT_CHARSET
    return HttpResponse(template.data, content_type=content_type)



# -*- coding: utf-8 -*-
# Copyright (C) 1998-2023 by the Free Software Foundation, Inc.
#
# This file is part of Postorius.
#
# Postorius is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Postorius is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Postorius.  If not, see <http://www.gnu.org/licenses/>.

"""Postorius view decorators."""
from functools import wraps

from django.core.exceptions import PermissionDenied

from postorius.auth.utils import set_list_access_props


def list_owner_required(fn):
    """Check if the logged in user is the list owner of the given list.
    Assumes that the request object is the first arg and that list_id
    is present in kwargs.
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        user = args[0].user
        list_id = kwargs['list_id']
        if not user.is_authenticated:
            raise PermissionDenied
        if user.is_superuser:
            return fn(*args, **kwargs)
        set_list_access_props(user, list_id, moderator=False)
        if user.is_list_owner:
            return fn(*args, **kwargs)
        else:
            raise PermissionDenied(f'User {user} is not Owner of {list_id}')

    return wrapper


def list_moderator_required(fn):
    """Check if the logged in user is a moderator of the given list.
    Assumes that the request object is the first arg and that list_id
    is present in kwargs.
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        user = args[0].user
        list_id = kwargs['list_id']
        if not user.is_authenticated:
            raise PermissionDenied
        if user.is_superuser:
            return fn(*args, **kwargs)
        set_list_access_props(user, list_id)
        if user.is_list_owner or user.is_list_moderator:
            return fn(*args, **kwargs)
        else:
            raise PermissionDenied(
                f'User {user} is not Owner or Moderator of {list_id}'
            )

    return wrapper


def superuser_required(fn):
    """Make sure that the logged in user is a superuser or otherwise raise
    PermissionDenied.
    Assumes the request object to be the first arg."""

    @wraps(fn)
    def wrapper(*args, **kwargs):
        user = args[0].user
        if not user.is_superuser:
            raise PermissionDenied
        return fn(*args, **kwargs)

    return wrapper



# -*- coding: utf-8 -*-
# Copyright (C) 1998-2023 by the Free Software Foundation, Inc.
#
# This file is part of Postorius.
#
# Postorius is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Postorius is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Postorius.  If not, see <http://www.gnu.org/licenses/>.


from urllib.error import HTTPError

from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import TemplateView

from django_mailman3.lib.mailman import get_mailman_client
from django_mailman3.lib.paginator import paginate

from postorius.auth.utils import set_list_access_props
from postorius.forms import AddBanForm
from postorius.models import List


class MailmanClientMixin(object):

    """Adds a mailmanclient.Client instance."""

    def client(self):
        if getattr(self, '_client', None) is None:
            self._client = get_mailman_client()
        return self._client


class MailingListView(TemplateView, MailmanClientMixin):

    """A generic view for everything based on a mailman.client
    list object.

    Sets self.mailing_list to list object if list_id is in **kwargs.
    """

    def get(self, request, *args, **kwargs):
        # This should be overridden by the subclass.
        return HttpResponse(status=405)

    def post(self, request, *args, **kwargs):
        # This should be overridden by the subclass.
        return HttpResponse(status=405)

    def _get_list(self, list_id, page):
        return List.objects.get_or_404(fqdn_listname=list_id)

    def dispatch(self, request, *args, **kwargs):
        # get the list object.
        if 'list_id' in kwargs:
            self.mailing_list = self._get_list(
                kwargs['list_id'], int(kwargs.get('page', 1))
            )
            set_list_access_props(request.user, self.mailing_list)
        # set the template
        if 'template' in kwargs:
            self.template = kwargs['template']
        return super(MailingListView, self).dispatch(request, *args, **kwargs)

    def _has_pending_unsub_req(self, email):
        """Check if there is a pending unsubscription request for email."""
        for req in self.mailing_list.get_requests(
            token_owner='moderator', request_type='unsubscription'
        ):
            if req.get('email') == email:
                return req
        return False


def bans_view(request, template, list_id=None):
    """Ban or unban email addresses.

    This is a reusable view which works for both global and list specific bans.
    Whether a MailingList ban is updated or a Global one depends on list_id
    being passed in.

    :list_id: MailingList Id if this is a List ban, None otherwise.

    """
    if list_id:
        m_list = List.objects.get_or_404(fqdn_listname=list_id)
        url = reverse('list_bans', args=[list_id])
        ban_list = m_list.bans
    else:
        ban_list = get_mailman_client().bans
        url = reverse('global_bans')
        m_list = None

    # Process form submission.
    if request.method == 'POST':
        if 'add' in request.POST:
            addban_form = AddBanForm(request.POST)
            if addban_form.is_valid():
                try:
                    ban_list.add(addban_form.cleaned_data['email'])
                    messages.success(
                        request,
                        _('The email {} has been banned.').format(
                            addban_form.cleaned_data['email']
                        ),
                    )
                except HTTPError as e:
                    messages.error(
                        request, _('An error occurred: %s') % e.reason
                    )
                except ValueError as e:
                    messages.error(request, _('Invalid data: %s') % e)
                return redirect(url)
        elif 'del' in request.POST:
            try:
                ban_list.remove(request.POST['email'])
                messages.success(
                    request,
                    _('The email {} has been un-banned').format(
                        request.POST['email']
                    ),
                )
            except HTTPError as e:
                messages.error(request, _('An error occurred: %s') % e.reason)
            except ValueError as e:
                messages.error(request, _('Invalid data: %s') % e)
            return redirect(url)
    else:
        addban_form = AddBanForm(initial=request.GET)
    banned_addresses = paginate(
        list(ban_list), request.GET.get('page'), request.GET.get('count')
    )

    context = {
        'addban_form': addban_form,
        'banned_addresses': banned_addresses,
    }

    if list_id:
        context['list'] = m_list

    return render(request, template, context)



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of Postorius.
#
# Postorius is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Postorius is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Postorius.  If not, see <http://www.gnu.org/licenses/>.


import logging

from django.utils import timezone

from postorius import __version__


logger = logging.getLogger(__name__)
# The day of the month which we celebrate as Mailman Day!
MAILMAN_DAY = 1


def postorius(request):
    """Add template variables to context."""
    return dict(
        POSTORIUS_VERSION=__version__,
        # Mailman Day is first of every month.
        mailman_day=(timezone.localtime(timezone.now()).day == MAILMAN_DAY),
    )
